import {  HvigorPlugin, HvigorNode, ModuleImpl, hvigor, TaskOutput } from '@ohos/hvigor';
import child_process from 'child_process'
import fs from 'fs'

// 获取当前hvigorNode节点对象
const HRouterPluginName = "hrouter"
const HRouterBootClazzName = "HRouterBoot.ets"
const PLUGIN_ID_NAME = "HRouterProcessor"
const Log = {
  i : function(...args) {
    console.log(`> hvigor [${PLUGIN_ID_NAME}]`, ...args)
  }
}
const ignoreFiles = [
  '.DS_Store'
]

/*
 * 检查当前的编译环境
 * */
function checkEnvNode(node:HvigorNode) {
  let hostProjPath = node.node._nodePath // 宿主工程文件目录
  let osModulePath = `${hostProjPath}/oh_modules/${HRouterPluginName}`
  if (!fs.existsSync(osModulePath)) {
    throw Error(`当前工程未集成 ${HRouterPluginName} 组件，请集成后，重新编译~`)
  }
}

/*
 * 寻找工程的主入口模块 entry 的 hap
 * */
function getEntryNodeBydNodeHvigor(node:HvigorNode) : (ModuleImpl | null) {
  let moduleImpl: ModuleImpl | null = null
  let map = (node.node._subProjects as Map<String, Object>)
  for (let items of map) {
    let path = `${items[1]._nodePath}/src/main/module.json5`;
    let type = JSON.parse(fs.readFileSync(path)).module.type;
    if (type == 'entry') {
      moduleImpl = items[1];
      break;
    }
  }
  return moduleImpl
}

/**
 * 创建临时目录
 * */
function createTemplateDirectory(files:List<object>) {
  let json = {}
  files.forEach((source)=>{
    let path = source["path"]
    let name = source["name"]
    if (source.forcedNew) {
      if (fs.existsSync(path)) fs.rmSync(path, { recursive: true })
    }

    if (!fs.existsSync(path)) {
      fs.mkdirSync(path, { recursive: true })
    }
    json[name] = path
    Log.i(`已经创建目录：${name}   ${path}`)
  })
  return json
}

/*
 * 备份文件目录
 * */
function backupFiles(sources:List<string>,  targetDir:string) {
  let json = {}
  sources.forEach((source)=>{
    let f = source.split("/").pop()
    let backupf = `__backup_${f}`
    if (!fs.existsSync(`${targetDir}/${backupf}`)) {
      Log.i(`备份：${source} 至：${targetDir}/${backupf}`)
      // 备份该文件，在编译完成后，还原该文件
      fs.copyFileSync(`${source}`, `${targetDir}/${backupf}`)
    } else {
      Log.i(`备份还原：${targetDir}/${backupf} 至：${source}`)
      fs.copyFileSync(`${targetDir}/${backupf}`, `${source}`)
    }
    json[f] = {
      source:source,
      backup:`${targetDir}/${backupf}`
    }
  })
  return json
}

/*
 * 开始安装插件辅助类
 * */
function installHRouterPlugin(node:HvigorNode, targetPath:string) {
  let hRouterPath = `${node.node._nodePath}/oh_modules/${HRouterPluginName}`
  let hRouterPagesRoot = `_hRouter/pages`
  let hRouterPageSourcePath = `${hRouterPath}/src/main/ets/pages`
  let jsonSrcArr = []
  if (fs.existsSync(hRouterPageSourcePath)) {
    let list = fs.readdirSync(hRouterPageSourcePath)
    if (list.length > 0) {
      list.forEach((fileName) => {
        if (ignoreFiles.indexOf(fileName)<0) {
          fs.copyFileSync(`${hRouterPageSourcePath}/${fileName}`, `${targetPath}/${fileName}`)
          jsonSrcArr.push(`${hRouterPagesRoot}/${fileName.split(".")[0]}`)
        }
      })
    }
  }
  return jsonSrcArr
}

function registerMainPagesJson(sourceFile:string, targetFile:string, concatSrcArr:List<string>) {
  // 注册路由
  let mainPageJsonSource = JSON.parse(fs.readFileSync(sourceFile));
  let srcArr = mainPageJsonSource.src;
  srcArr = srcArr.concat(concatSrcArr)
  mainPageJsonSource.src = srcArr;
  fs.writeFileSync(targetFile, JSON.stringify(mainPageJsonSource))
}

/**
 * 递归找到对应的文件名称
 * */
function recursionSearchFile(src, fileName) {
  try {
    let ls = fs.readdirSync(src)
    for (let f of ls) {
      if (ignoreFiles.indexOf(f) >= 0) continue;
      let fstate = fs.statSync(`${src}/${f}`)
      if (fstate.isDirectory()) {
        let lastf = recursionSearchFile(`${src}/${f}`, fileName)
        if (lastf != null) {
          return lastf
        }
      } else {
        if (f.split(".")[0] == fileName) {
          return `${src}/${f}`
        }
      }
    }
  } catch (e) {
    console.log(e)
  }
}
/**
 * 递归搜索依赖模块中的子模块
 * */
function recursionOHModules(src, fn) {
  try {
    let ls = fs.readdirSync(src)
    for (let f of ls) {
      if (ignoreFiles.indexOf(f) >= 0) continue;
      recursionOHModulesSubFiles(`${src}/${f}/src/main/ets`, (fileName)=>{
        fn && fn(f, fileName)
      })
    }
  } catch (e) {
    console.log(e)
  }
}
/**
 * 递归检索依赖模块中的子文件
 * */
function recursionOHModulesSubFiles(src, fn) {
  if (fs.existsSync(src)) {
    let ls = fs.readdirSync(src)
    for (let f of ls) {
      if (ignoreFiles.indexOf(f) >= 0) continue;
      let fstate = fs.statSync(`${src}/${f}`)
      if (fstate.isDirectory()) {
        recursionOHModulesSubFiles(`${src}/${f}`, fn)
      } else {
        let path = `${src}/${f}`
        if ((fs.readFileSync(path).toString().indexOf("@HRoute") >= 0) ||
              fs.readFileSync(path).toString().indexOf("@HPluginRoute") >= 0) {
          fn && fn(f, path)
        }
      }
    }
  }
}
/**
 * 计算从 src1 引用 src2 的相对路径
 * */
function absoluteSrc1RefSrc2(src1:string, src2:string) {
  /**
   * TODO:
   * 例如：
   * src1 = /A/B/C/D.ets，src2 = /A/B/C/F/E.ets，结果 ./F/E.ets
   * src1 = /A/B/G.ets, src2 = /A/B/C/F/E.ets, 结果 ./C/F/E.ets
   * src1 = /A/B/C/H/G.ets, src2 = /A/B/C/F/E.ets, 结果 ../F/E.ets
   * src1 = /A/B/C/F/G.ets, src2 = /A/B/C/F/E.ets, 结果 ./E.ets
   * */
  // 将路径分割成数组
  const parts1 = src1.split('/');
  const parts2 = src2.split('/');

  // 找到共同前缀的长度
  let commonLength = 0;
  for (let i = 0; i < Math.min(parts1.length, parts2.length); i++) {
    if (parts1[i] === parts2[i]) {
      commonLength++;
    } else {
      break;
    }
  }

  // 计算从 src1 到共同前缀的回溯路径
  const backtrack = Array(parts1.length - commonLength - 1).fill('..'); // -1 为最后的文件名或类名

  // 计算从共同前缀到 src2 的路径
  const forward = parts2.slice(commonLength);

  // 合并这两部分来形成相对路径
  let relativePath = [...backtrack, ...forward].join('/');
  if (!relativePath.startsWith('../')) {
    relativePath = './' + relativePath; // 如果路径从当前目录开始，添加 './'
  }

  return relativePath.substr(0, relativePath.length - 4);
}

function watchCompileFinish(node: HvigorNode, entryModuleImpl:ModuleImpl){
  // TODO:将备份的文件还原，让开发者无感编译
  let child = child_process.spawn("node", [`${__dirname}/ProcessWatchCompile.js`, node.node._nodePath, process.pid, entryModuleImpl._nodePath], {
    detached: true,
    stdio: 'ignore'
  })
  child.unref();
}

// 实现自定义插件
export function HRouterProcessor(debug:boolean = false): HvigorPlugin {

  return {
    pluginId: PLUGIN_ID_NAME,
    apply(node: HvigorNode) {


      Log.i("开始组件预处理...")
      // 检查项目有没有依赖 HRouter 模块，没有的话，直接编译报错
      Log.i("检查编译环境")
      checkEnvNode(node);
      Log.i("开始寻找项目的 entry 模块入口")
      let entryModuleImpl = getEntryNodeBydNodeHvigor(node)
      if (!debug) {
        watchCompileFinish(node, entryModuleImpl)
      }
      Log.i("工程的 entry 模块：", entryModuleImpl._nodePath)
      let tmpJson = createTemplateDirectory([
        {
          path:`${entryModuleImpl._nodePath}/__tmp`,
          name:"__tmp"
        },
        {
          path:`${entryModuleImpl._nodePath}/src/main/ets/_hRouter/pages`,
          name:"__hRouter",
          forcedNew:true,
        }
      ])

      Log.i(`开始备份文件`)
      let entryRootPath = entryModuleImpl._nodePath;
      let moduleJson5 = JSON.parse(fs.readFileSync(`${entryRootPath}/src/main/module.json5`))
      let entryMainEl = moduleJson5.module.mainElement;
      let enryMainElSource = recursionSearchFile(`${entryRootPath}/src/main/ets`, entryMainEl)
      let backupJson = backupFiles([
        `${entryRootPath}/src/main/resources/base/profile/main_pages.json`,
        enryMainElSource
      ], tmpJson.__tmp)
      Log.i(`结束备份文件`)

      // 把 pages 文件移动到，宿主目录中
      Log.i(`开始安装 HRouter 插件辅助类`)
      let jsonSrcArr = installHRouterPlugin(node, tmpJson.__hRouter)
      Log.i(`已安装类文件：${jsonSrcArr}`)
      let backUpFileJson = backupJson['main_pages.json']
      fs.writeFileSync(`${tmpJson.__tmp}/__backupjson`, JSON.stringify(backupJson))
      registerMainPagesJson(
        backUpFileJson.backup,
        backUpFileJson.source,
        jsonSrcArr
      )
      Log.i(`结束安装 HRouter 插件辅助类`)


      // 创建引导文件
      let bootHRouterFilePath = `${entryRootPath}/src/main/ets/_hRouter/${HRouterBootClazzName}`

      // 加入引导类
      let enryMainElSrc = backupJson["EntryAbility.ets"].source
      let srcCnt = fs.readFileSync(enryMainElSrc).toString()
      let srcList_b = srcCnt.split("export default class")
      srcList_b[0] += `import { HEntryRoute } from \'${HRouterPluginName}\';\r\n\r\n @HEntryRoute(HRouterBoot)\r\n`
      srcCnt = srcList_b.join("export default class")
      srcCnt = `import * as HRouterBoot from '${absoluteSrc1RefSrc2(enryMainElSrc, bootHRouterFilePath)}';\r\n${srcCnt}`
      fs.writeFileSync(enryMainElSrc, srcCnt)
      Log.i(`已增加引导类：${HRouterBootClazzName}`)

      // 扫描项目依赖中所有的类文件
      let contentBootRouter = ''
      let exportStr = ""
      let ohModulesPath = `${node.node._nodePath}/oh_modules`
      recursionOHModules(ohModulesPath, (moduleName, pageName)=>{
        let fname = pageName.split('.')[0]
        contentBootRouter += `import { ${fname} } from '${moduleName}';\r\n`
        exportStr += `${fname},\r\n`
      })
      recursionOHModulesSubFiles(`${entryRootPath}/src/main/ets`, function(fileName, absPath){
        console.log(fileName, absPath)
        let f = fileName.split(".")[0]
        contentBootRouter += `import { ${f} } from '${absoluteSrc1RefSrc2(bootHRouterFilePath, absPath)}';\r\n`
        exportStr += `${f},\r\n`
      })

      contentBootRouter += `\r\n\r\nexport  { \r\n${exportStr} } `
      fs.writeFileSync(bootHRouterFilePath, contentBootRouter)


      Log.i("结束组件预处理...")
    }
  }
}
